// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Program wcfs provides filesystem server with file data backed by wendelin.core arrays.
//
// Intro
//
// Each wendelin.core array (ZBigArray) is actually a linear file (ZBigFile)
// and array metadata like dtype, shape and strides associated with it. This
// program exposes as files only ZBigFile data and leaves rest of
// array-specific handling to clients. Every ZBigFile is exposed as one separate
// file that represents whole ZBigFile's data.
//
// For a client, the primary way to access a bigfile should be to mmap
// bigfile/<bigfileX>/head/data which represents always latest bigfile data.
// Clients that want to get isolation guarantee should subscribe for
// invalidations and re-mmap invalidated regions to file with pinned bigfile revision for
// the duration of their transaction. See "Invalidation protocol" for details.
//
// In the usual situation when bigfiles are big, and there are O(1)/δt updates,
// there should be no need for any cache besides shared kernel cache of latest
// bigfile data.
//
//
// Filesystem organization
//
// Top-level structure of provided filesystem is as follows:
//
//	bigfile/
//		<oid(bigfile1)>/
//			...
//		<oid(bigfile2)>/
//			...
//		...
//
// where for a bigfileX there is bigfile/<oid(bigfileX)>/ directory, with
// oid(bigfileX) being ZODB object-id of corresponding ZBigFile object formatted with %016x.
//
// Each bigfileX/ has the following structure:
//
//	bigfile/<bigfileX>/
//		head/		; latest bigfile revision
//			...
//		@<tid1>/	; bigfile revision as of transaction <tidX>
//			...
//		@<tid2>/
//			...
//		...
//
// where head/ represents latest bigfile as stored in upstream ZODB, and
// @<tidX>/ represents bigfile as of transaction <tidX>.
//
// head/ has the following structure:
//
//	bigfile/<bigfileX>/head/
//		data		; latest bigfile data
//		at		; data is bigfile view as of this ZODB transaction
//		invalidations	; channel that describes invalidated data regions
//
// where /data represents latest bigfile data as stored in upstream ZODB. As
// there can be some lag receiving updates from the database, /at describes
// precisely ZODB state for which bigfile data is currently exposed. Whenever
// bigfile data is changed in upstream ZODB, information about the changes is
// first propagated to /invalidations, and only after that /data is
// updated. See "Invalidation protocol" for details.
//
// @<tidX>/ has the following structure:
//
//	bigfile/<bigfileX>/@<tidX>/
//		data		; bigfile data as of transaction <tidX>
//
// where /data represents bigfile data as of transaction <tidX>.
//
// bigfile/<bigfileX>/ should be created by client via mkdir. Unless explicitly
// created bigfile/<bigfileX>/ are not automatically visible in wcfs
// filesystem. Similarly bigfile/<bigfileX>/@<tidX>/ should be too created by
// client.
//
//
// Invalidation protocol
//
// XXX invalidations will be done via ptrace because we need them to be
// synchronous (see "wcfs organization")
//
// In order to support isolation wcfs implements invalidation protocol that
// must be cooperatively followed by both wcfs and client.
//
// First, before client wants to mmap bigfile, it opens
// bigfile/<bigfileX>/head/invalidations and tells wcfs through it for which
// ZODB state it wants to get bigfile view. The server in turn reports for
// which ZODB state head/data is current, δ describing changed bigfile region
// between those revisions, or "wait" flag if server state is earlier compared
// to what client wants:
//
//	C: want <Cat>
//	S: have <Sat>, wait		; Sat < Cat
//	S: have <Sat>, δR(Cat,Sat)	; Sat ≥ Cat
//
// If server reply was "wait" the client does nothing and waits for next server
// message which must come without "wait" flag set. When client receives have
// message with δR(Cat,Sat) it has the guarantee from wcfs that head/data
// content is for Sat ZODB revision and won't change until client sends ack
// back to the server. The client in turn now can mmap head/data and
// @<Cat>/data to get bigfile view as of Cat:
//
//	mmap(bigfile/<bigfileX>/head/data)
//	mmap(bigfile/<bigfileX>/@<Cat>/data, δR(Cat,Sat), MAP_FIXED)  # mmaped at addresses corresponding to δR(Cat,Sat)
//
// When client completes its initial mmapping it sends ack back to the server:
//
//	C: ack
//
// From now on the server will be processing updates to bigfile coming from
// ZODB as follows:
//
//
// The filesystem server itself receives information about changed data
// from ZODB server through regular ZODB invalidation channel (as it is ZODB
// client itself). Then, before actually updating bigfile/<bigfileX>/head/data
// content in changed part, it notifies through bigfile/<bigfileX>/head/invalidations
// to clients that had opened this file (separately to each client) about the changes:
//
//	S: have <Sat>, δR(Sat_prev, Sat)
//
// where Sat_prev is ZODB revision last reported to client for this bigfile,
// and waits until they all confirm that changed file part can be updated in
// global OS cache.
//
// The client in turn can now re-mmap invalidated regions to bigfile@Cat
//
//	# mmapped at addresses corresponding to δR(Sat_prev, Sat)
//	mmap(bigfile/<bigfileX>/@<Cat>/data, δR(Sat_prev, Sat), MAP_FIXED)
//
// and must send ack back to the server when it is done:
//
//	C: ack
//
// When clients are done with bigfile/<bigfileX>/@<Cat>/data (i.e. Cat
// transaction ends and array is unmapped), the server sees number of opened
// files to bigfile/<bigfileX>/@<Cat>/data drops to zero, and automatically
// destroys bigfile/<bigfileX>/@<Cat>/ directory after reasonable timeout.
//
//
// Protection against slow or faulty clients
//
// If a client, on purpose or due to a bug or being stopped, is slow to
// respond with ack to invalidation notification, it creates a problem because
// head/data updates will be blocked and thus all other clients that try to
// work with current data will get stuck.
//
// To avoid this problem it should be possible for wcfs to stop a client with
// ptrace and change its address space in a style similar to e.g.
// VirtualAllocEx on Windows. Here is hacky example how this could be done on Linux:
//
// https://gist.github.com/rofl0r/1073739/63f0f788a4923e26fcf743dd9a8411d4916f0ac0
//
// This way there should be no possibility for a client to block wcfs
// indefinitely waiting for client's ack.
//
// Similarly for initial mmapings client could first mmap head/data, then open
// head/invalidations and tell the server that it wants Cat revision, with
// the server then remmaping blocks to get to Cat state via ptrace.
//
// However for simplicity the plan is to go first without ptrace and just kill
// a slow client on, say 30 seconds, timeout.
//
//
// Writes
//
// As each bigfile is represented by 1 synthetic file, there can be several
// write schemes:
//
// 1. mmap(MAP_PRIVATE) + writeout by client
//
// In this scheme bigfile data is mmapped in MAP_PRIVATE mode, so that local
// user changes are not automatically propagated back to the file. When there
// is a need to commit, client investigates via some OS mechanism, e.g.
// /proc/self/pagemap or something similar, which pages of this mapping it
// modified. Knowing this it knows which data it dirtied and so can write this
// data back to ZODB itself, without filesystem server providing write support.
//
// 2. mmap(MAP_SHARED, PROT_READ) + write-tracking & writeout by client
//
// In this scheme bigfile data is mmaped in MAP_SHARED mode with read-only pages
// protection. Then whenever write fault occurs, client allocates RAM from
// shmfs, copies faulted page to it, and then mmaps RAM page with RW protection
// in place of original bigfile page. Writeout implementation should be similar
// to "1", only here client already knows the pages it dirtied, and this way
// there is no need to consult /proc/self/pagemap.
//
// The advantage of this scheme over mmap(MAP_PRIVATE) is that in case
// there are several in-process mappings of the same bigfile with overlapping
// in-file ranges, changes in one mapping will be visible in another mapping.
// Contrary: whenever a MAP_PRIVATE mapping is modified, the kernel COWs
// faulted page into a page completely private to this mapping, so that other
// MAP_PRIVATE mappings of this file, including ones created from the same
// process, do not see changes made to the first mapping.
//
// Since wendelin.core needs to provide coherency in between different slices
// of the same array, this is the mode wendelin.core actually uses.
//
// 3. write to wcfs
//
// XXX we later could implement "write-directly" mode where clients would write
// data directly into the file.
package main

// Wcfs organization
//
// Wcfs is a ZODB client that translates ZODB objects into OS files as would
// non-wcfs wendelin.core do for a ZBigFile. It is organized as follows:
//
// - 1 ZODB connection for "latest data" for whole filesystem (zconn).
// - head/data of all bigfiles represent state as of zconn.At .
// - for */head/data the following invariant is maintained:
//
//	#blk ∈ file cache    =>    ZBlk(#blk) + all BTree/Bucket that lead to it  ∈ zconn cache
//	                           (ZBlk* in ghost state)
//
//   The invariant helps on invalidation: if we see a changed oid, and
//   zconn.cache.lookup(oid) = ø -> we know we don't have to invalidate OS
//   cache for any part of any file (even if oid relates to a file block - that
//   block is not cached and will trigger ZODB load on file read).
//
//   Currently we maintain this invariant by simply never evicting LOBTree/LOBucket
//   objects from ZODB Connection cache (LOBucket keeps references to ZBlk* and
//   so ZBlk* also stay in cache in ghost form). In the future we may want to
//   try to synchronize to kernel freeing its pagecache pages.
//
// - when we receive an invalidation message from zstor - we process it and
//   propagate invalidations to OS file cache of */head/data:
//
//	invalidation message: (tid↑, []oid)
//
//   1. zconn.cache.lookup(oid)
//   2. ø: nothing to do - see invariant ^^^.
//   3. obj found:
//
//	- ZBlk*		-> file/#blk
//	- BTree/Bucket	-> δ(BTree)  -> file/[]#blk
//
//     in the end after processing all []oid from invalidation message we have
//
//	  [] of file/[]#blk
//
//     that describes which file(s) parts needs to be invalidated.
//
//   4. for all file/blk to invalidate we do:
//
//	- try to retrieve file/head/data[blk] from OS file cache;
//	- if retrieved successfully -> store retrieved data back into OS file
//	  cache for file/@<rev>/data[blk], where
//
//	    rev = max(δFtail.by(#blk)) || zconn.at	; see below about δFtail
//
//	- invalidate file/head/data[blk] in OS file cache.
//
//	This preserves previous data in OS file cache in case it will be needed
//	by not-yet-uptodate clients, and makes sure file read of head/data[blk]
//	won't be served from OS file cache and instead will trigger a FUSE read
//	request to wcfs.
//
// - for every file δFtail invalidation info about head/data is maintained:
//
//	- tail: of [](rev↑, []#blk)
//	- by:   {} #blk -> []rev↑ in tail
//
//      δFtail.tail describes invalidations to file we learned from ZODB invalidation.
//      δFtail.by   allows to quickly lookup information by #blk.
//
//      min(rev) in δFtail is min(@at) at which head/data is currently mmapped (see below).
//
// - when we receive a FUSE read(#blk) request to a file/head/data we process it as follows:
//
//   1. load blkdata for head/data[blk] @zconn.at .
//
//      while loading this also gives upper bound estimate of when the block
//      was last changed:
//
//	  rev(blk) ≤ max(_.serial for _ in (ZBlk(#blk), all BTree/Bucket that lead to ZBlk))
//
//	it is not exact because BTree/Bucket can change (e.g. rebalance)
//	but still point to the same k->ZBlk.
//
//	we also use file.δFtail to find either exact blk revision:
//
//	  rev(blk) = max(file.δFtail.by(#blk) -> []rev↑)
//
//	or another upper bound if #blk ∉ δFtail:
//
//	  rev(blk) ≤ min(rev) is δFtail		; #blk ∉ δFtail
//
//
//	below rev'(blk) is min(of the estimates found):
//
//	  rev(blk) ≤ rev'(blk)		rev'(blk) = min(^^^)
//
//
//   2. for all client/addr@at mmappings of file/head/data:
//
//	- rev'(blk) ≤ at: -> do nothing
//	- rev'(blk) > at:
//	  - if blk ∈ mmapping.pinned -> do nothing
//	  - rev = max(δFtail.by(#blk) : _ ≤ at)	|| at
//	  - client.remmap(addr[blk], file/@rev/data)
//	  - mmapping.pinned += blk
//
//	remmapping is done synchronously via ptrace.
//	XXX via running wcfs-trusted code wcfs injects into clients.
//	FIXME ptrace won't work when client thread is blocked under pagefault or syscall.
//
//	in order to support remmapping for each file/head/data
//
//	  [] of mmapping{client/addr/@at↑, pinned}
//
//	is maintained.
//
//   3. blkdata is returned to kernel.
//
//   Thus a client that wants latest data on pagefault will get latest data,
//   and a client that wants @rev data will get @rev data, even if it was this
//   "old" client that triggered the pagefault.
//
//
// XXX mmap(@at) open
//
// XXX zconn(s) for historical state
// XXX serving read from @<rev>/data
//
// XXX(integrate place=?) ZData - ne need to keep track -> ZBlk1 is always
// marked as changed on blk data change.
//
// ----------------------------------------
//
// δ(BTree) notes
//
//
// input: BTree, (@new, []oid)  ->  find out δ(BTree) i.e. {-k(v), +k'(v'), ...}
//
// - oid ∈ Bucket
// - oid ∈ BTree
//
// Bucket:
//
//	old = {k  -> v}
//	new = {k' -> v'}
//
//	Δ = -k(v), +k(v), ...
//
// => for all buckets
//
//	Δ accumulates to []δk(v)[n+,n-]  n+ ∈ {0,1}, n- ∈ {0,1}, if n+=n- - cancel
//
//
// BTree:
//
//	old = {k  -> B}   or {k  -> T}
//	new = {k' -> B'}  or {k' -> T'}
//
//	Δ = -k(B), +k(B), -k(T), +K(T), ...
//
// we translate (in top-down order):
//
//	k(B) -> {} of k(v)
//	k(T) -> {} of k(B) -> {} of k(v)
//
// which gives
//
//	Δ = k(v), +k(v), ...
//
// i.e. exactly as for buckets and it accumulates to global Δ.
//
// The globally-accumulated Δ is the answer for δ(BTree, (@new, []oid))
//
// XXX -> internal/btreediff ?
//
// δ(BTree) in wcfs context:
//
// . -k(blk) -> invalidata #blk
// . +k(blk) -> invalidate #blk (e.g. if blk was previously read as hold)
//
//
// ----------------------------------------
//
// - XXX(kill) 1 ZODB connection per 1 bigfile (each bigfile can be at its different @at,
//   because invalidations for different bigfiles can be processed with different
//   timings depending on clients). No harm here as different bigfiles use
//   completely different ZODB BTree and data objects.
//
//   For every ZODB connection a dedicated read-only transaction is maintained.
//
//
// Notes on OS pagecache control:
//
// the cache of snapshotted bigfile can be pre-made hot, if invalidated region
// was already in pagecache of head/data:
//
// - we can retrieve a region from pagecache of head/data with FUSE_NOTIFY_RETRIEVE.
// - we can store that retrieved data into pagecache region of @<tidX>/ with FUSE_NOTIFY_STORE.
// - we can invalidate a region from pagecache of head/data with FUSE_NOTIFY_INVAL_INODE.
//
// we have to disable FUSE_AUTO_INVAL_DATA to tell the kernel we are fully
// responsible for invalidating pagecache. If we don't, the kernel will be
// clearing whole cache of head/data on e.g. its mtime change.
//
// XXX FUSE_AUTO_INVAL_DATA does not fully prevent kernel from automatically
// invalidating pagecache - e.g. it will invalidate whole cache on file size changes:
//
// https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/fs/fuse/inode.c?id=e0bc833d10#n233
//
// we can currently workaround it with using writeback mode (see !is_wb in the
// link above), but better we have proper FUSE flag for filesystem server to
// tell the kernel it is fully responsible for invalidating pagecache.

import (
	"context"
	"flag"
	stdlog "log"
	"os"
	"strings"
	"sync"
	"syscall"

	log "github.com/golang/glog"
	"golang.org/x/sync/errgroup"

	"lab.nexedi.com/kirr/go123/xcontext"
	"lab.nexedi.com/kirr/go123/xerr"

	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
	_ "lab.nexedi.com/kirr/neo/go/zodb/wks"

	"github.com/hanwen/go-fuse/fuse"
	"github.com/hanwen/go-fuse/fuse/nodefs"
	"github.com/pkg/errors"
)

// /bigfile/	- served by BigFileRoot.
type BigFileRoot struct {
	nodefs.Node
	zstor zodb.IStorage

	// {} oid -> <bigfileX>/
	mu  sync.Mutex
	tab map[zodb.Oid]*BigFileDir
}

// /bigfile/<bigfileX>/	- served by BigFileDir.
type BigFileDir struct {
	nodefs.Node
	oid zodb.Oid // oid of ZBigFile

	// ZODB DB handle for this bigfile.
	// keeps cache of connections for both head/ and @<rev>/ accesses.
	zdb *zodb.DB

	// head/ is implicitly linked to by fs

	// {} rev -> @<rev>/ bigfile snapshot
	mu     sync.Mutex
	revTab map[zodb.Tid]*BigFileRev
}

// /bigfile/<bigfileX>/(head|<rev>)/	- served by BigFileRev.
type BigFileRev struct {
	nodefs.Node
	// data, at, invalidations, etc - all implicitly linked to by fs
}

// /bigfile/<bigfileX>/(head|<rev>)/*	- internally served by BigFile.
type BigFile struct {
	// current read-only transaction under which we access ZODB data
	txnCtx context.Context // XXX -> better directly store txn

	// connection via which ZODB objects for this bigfile are accessed
	// XXX do we need to keep it here explicitly?
	zconn	*zodb.Connection

	// ZBigFile top-level object. Kept activated during lifetime of current transaction.
	zbf	*ZBigFile

	// zbf.Size(). It is constant during liftime of current transaction.
	zbfSize int64

	// TODO
	// lastChange	zodb.Tid // last change to whole bigfile as of .zconn.At view
}

// /bigfile/<bigfileX>/(head|<rev>)/data	- served by BigFileData.
type BigFileData struct {
	nodefs.Node

	bigfile *BigFile

	// inflight loadings of ZBigFile from ZODB.
	// successfull load results are kept here until blkdata is put into OS pagecache.
	//
	// XXX -> BigFile ?
	loadMu  sync.Mutex
	loading map[int64]*blkLoadState // #blk -> {... blkdata}
}

// blkLoadState represents a ZBlk load state/result.
//
// when !ready the loading is in progress.
// when ready the loading has been completed.
type blkLoadState struct {
	ready chan struct{}

	blkdata []byte
	err     error
}


// ----------------------------------------

// /bigfile -> Mkdir receives client request to create /bigfile/<bigfileX>.
//
// It creates <bigfileX>/head/* along the way.
func (bfroot *BigFileRoot) Mkdir(name string, mode uint32, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	inode, err := bfroot.mkdir(name, fctx) // XXX ok to ignore mode?
	return inode, err2LogStatus(err)

}

func (bfroot *BigFileRoot) mkdir(name string, fctx *fuse.Context) (_ *nodefs.Inode, err error) {
	defer xerr.Contextf(&err, "/bigfile: mkdir %q", name)

	oid, err := zodb.ParseOid(name)
	if err != nil {
		return nil, eINVALf("not oid")
	}

	// check to see if dir(oid) is already there
	bfroot.mu.Lock()
	_, already := bfroot.tab[oid]
	bfroot.mu.Unlock()

	if already {
		return nil, syscall.EEXIST
	}

	// not there - without bfroot lock proceed to open BigFile from ZODB
	zdb := zodb.NewDB(bfroot.zstor)
	bf, err := bigopen(asctx(fctx), zdb, oid, &zodb.ConnOptions{})
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			bf.Close()
		}
	}()

	// relock bfroot and either mkdir or EEXIST if the directory was maybe
	// simultanously created while we were not holding bfroot.mu
	bfroot.mu.Lock()
	_, already = bfroot.tab[oid]
	if already {
		bfroot.mu.Unlock()
		return nil, syscall.EEXIST
	}

	bfdir := &BigFileDir{
		Node:   nodefs.NewDefaultNode(),
		oid:    oid,
		zdb:    zdb,
		revTab: make(map[zodb.Tid]*BigFileRev),
	}

	bfhead := &BigFileRev{
		Node: nodefs.NewDefaultNode(),
	}

	bfdata := &BigFileData{
		Node:    nodefs.NewDefaultNode(),
		bigfile: bf,
		loading: make(map[int64]*blkLoadState),
	}

	bfroot.tab[oid] = bfdir
	bfroot.mu.Unlock()

	// mkdir takes filesystem treeLock - do it outside bfroot.mu
	mkdir(bfroot, name, bfdir)
	mkdir(bfdir, "head", bfhead)
	mkfile(bfhead, "data", bfdata)
	mkfile(bfhead, "at", NewSmallFile(bf.readAt))	// TODO mtime(at) = tidtime(at)
	// XXX mkfile(bh, "invalidations", bh.inv)

	return bfdir.Inode(), nil
}

// XXX do we need to support rmdir? (probably no)

// /bigfile/<bigfileX> -> Mkdir receives client request to create @<tid>/.
func (bfdir *BigFileDir) Mkdir(name string, mode uint32, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	inode, err := bfdir.mkdir(name, fctx) // XXX ok to ignore mode?
	return inode, err2LogStatus(err)
}

func (bfdir *BigFileDir) mkdir(name string, fctx *fuse.Context) (_ *nodefs.Inode, err error) {
	defer xerr.Contextf(&err, "/bigfile/%s: mkdir %q", bfdir.oid, name)

	var tid zodb.Tid
	ok := false

	if strings.HasPrefix(name, "@") {
		tid, err = zodb.ParseTid(name[1:])
		ok = (err == nil)
	}
	if !ok {
		return nil, eINVALf("not @tid")
	}

	// check to see if dir(tid) is already there
	bfdir.mu.Lock()
	_, already := bfdir.revTab[tid]
	bfdir.mu.Unlock()

	if already {
		return nil, syscall.EEXIST
	}

	// not there - without bfdir lock proceed to open BigFile @tid view of ZODB
	bf, err := bigopen(asctx(fctx), bfdir.zdb, bfdir.oid, &zodb.ConnOptions{
		At: tid,
	})
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			bf.Close()
		}
	}()

	// relock bfdir and either mkdir or EEXIST if the directory was maybe
	// simultanously created while we were not holding bfroot.mu
	bfdir.mu.Lock()
	_, already = bfdir.revTab[tid]
	if already {
		bfdir.mu.Unlock()
		return nil, syscall.EEXIST
	}

	bfrev := &BigFileRev{
		Node: nodefs.NewDefaultNode(),
	}

	revdata := &BigFileData{
		Node:    nodefs.NewDefaultNode(),
		bigfile: bf,
		loading: make(map[int64]*blkLoadState),
	}

	bfdir.revTab[tid] = bfrev
	bfdir.mu.Unlock()

	// mkdir takes filesystem treeLock - do it outside bfroot.mu
	mkdir(bfdir, name, bfrev)
	mkfile(bfrev, "data", revdata)

	return bfrev.Inode(), nil
}


// bigopen opens BigFile corresponding to oid and zopt.
//
// A new read-only transaction is opened.
// A new ZODB connection is opened via zdb.
// A ZBigFile corresponding to oid is activated and statted.
//
// The whole result is returned as BigFile.
func bigopen(ctx context.Context, zdb *zodb.DB, oid zodb.Oid, zopt *zodb.ConnOptions) (_ *BigFile, err error) {
	defer xerr.Contextf(&err, "bigopen %s %s", oid, zopt)

	// create new read-only transaction for this bigfile
	txn, txnCtx := transaction.New(context.Background())
	defer func() {
		if err != nil {
			txn.Abort()
		}
	}()


	// create new DB/Connection for this bigfile open

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(ctx, txnCtx)
	defer cancel()

	zconn, err := zdb.Open(ctx, zopt)
	if err != nil {
		return nil, err
	}

	xzbf, err := zconn.Get(ctx, oid)
	if err != nil {
		switch errors.Cause(err).(type) {
		case *zodb.NoObjectError:
			return nil, eINVAL(err)
		case *zodb.NoDataError:
			return nil, eINVAL(err) // XXX what to do if it was existing and got deleted?
		default:
			return nil, err
		}
	}

	zbf, ok := xzbf.(*ZBigFile)
	if !ok {
		return nil, eINVALf("%s is not a ZBigFile", typeOf(xzbf))
	}

	// activate ZBigFile and keep it this way
	err = zbf.PActivate(ctx)
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			zbf.PDeactivate()
		}
	}()

	zbfSize, err := zbf.Size(ctx)
	if err != nil {
		return nil, err
	}

	return &BigFile{
		txnCtx:  txnCtx,
		zconn:   zconn,
		zbf:     zbf,
		zbfSize: zbfSize,
	}, nil
}

// Close release all resources of BigFile.
func (bf *BigFile) Close() error {
	bf.zbf.PDeactivate()
	bf.zbf = nil

	transaction.Current(bf.txnCtx).Abort()
	bf.zconn = nil

	return nil
}

// /bigfile/<bigfileX>/head/data -> Getattr serves stat.
func (bfdata *BigFileData) GetAttr(out *fuse.Attr, _ nodefs.File, fctx *fuse.Context) fuse.Status {
	// XXX locking

	bf := bfdata.bigfile

	out.Mode = fuse.S_IFREG | 0444
	out.Size = uint64(bf.zbfSize)
	// .Blocks
	// .Blksize

	// FIXME lastChange should cover all bigfile data, not only ZBigFile itself
	//mtime := &bfdata.lastChange.Time().Time
	lastChange := bf.zbf.PSerial()

	mtime := lastChange.Time().Time
	out.SetTimes(/*atime=*/nil, /*mtime=*/&mtime, /*ctime=*/&mtime)

	return fuse.OK
}


// /bigfile/<bigfileX>/head/data -> Read serves reading bigfile data.
func (bfdata *BigFileData) Read(_ nodefs.File, dest []byte, off int64, fctx *fuse.Context) (fuse.ReadResult, fuse.Status) {
	// XXX locking

	bf := bfdata.bigfile
	zbf := bf.zbf

	// cap read request to file size
	end := off + int64(len(dest))		// XXX overflow?
	if end > bf.zbfSize {
		end = bf.zbfSize
	}
	if end <= off {
		// XXX off >= size -> EINVAL? (but when size=0 kernel issues e.g. [0 +4K) read)
		return fuse.ReadResultData(nil), fuse.OK
	}

	// widen read request to be aligned with blksize granularity
	// (we can load only whole ZBlk* blocks)
	aoff := off - (off % zbf.blksize)
	aend := end
	if re := end % zbf.blksize; re != 0 {
		aend += zbf.blksize - re
	}
	dest = make([]byte, aend - aoff) // ~> [aoff:aend) in file

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(asctx(fctx), bf.txnCtx)
	defer cancel()

	// read/load all block(s) in parallel
	wg, ctx := errgroup.WithContext(ctx)
	for blkoff := aoff; blkoff < aend; blkoff += zbf.blksize {
		blkoff := blkoff
		blk := blkoff / zbf.blksize
		wg.Go(func() error {
			δ := blkoff-aoff // blk position in dest
			//log.Infof("readBlk #%d dest[%d:+%d]", blk, δ, zbf.blksize)
			return bfdata.readBlk(ctx, blk, dest[δ:δ+zbf.blksize])
		})
	}

	err := wg.Wait()
	if err != nil {
		log.Errorf("%s", err)	// XXX + /bigfile/XXX: read [a,b): -> ...
		return nil, fuse.EIO
	}

	return fuse.ReadResultData(dest[off-aoff:end-aoff]), fuse.OK
}

// readBlk serves Read to read 1 ZBlk #blk into destination buffer.
//
// len(dest) == blksize.
func (bfdata *BigFileData) readBlk(ctx context.Context, blk int64, dest []byte) error {
	// XXX errctx?

	// check if someone else is already loading this block
	bfdata.loadMu.Lock()
	loading, already := bfdata.loading[blk]
	if !already {
		loading = &blkLoadState{
			ready:   make(chan struct{}),
		}
		bfdata.loading[blk] = loading
	}
	bfdata.loadMu.Unlock()

	// if it is already loading - just wait for it
	if already {
		select {
		case <-ctx.Done():
			return ctx.Err()

		case <-loading.ready:
			if loading.err == nil {
				copy(dest, loading.blkdata)
			}
			return loading.err
		}
	}

	// noone was loading - we became reponsible to load this block

	zbf := bfdata.bigfile.zbf
	blkdata, err := zbf.LoadBlk(ctx, blk)
	loading.blkdata = blkdata
	loading.err = err
	close(loading.ready)

	// data loaded with error - cleanup .loading
	if loading.err != nil {
		bfdata.loadMu.Lock()
		delete(bfdata.loading, blk)
		bfdata.loadMu.Unlock()
		return err
	}

	// data loaded ok
	copy(dest, blkdata)

	// store to kernel pagecache whole block that we've just loaded from database.
	// This way, even if the user currently requested to read only small portion from it,
	// it will prevent next e.g. consecutive user read request to again hit
	// the DB, and instead will be served by kernel from its pagecache.
	//
	// We cannot do this directly from reading goroutine - while reading
	// kernel FUSE is holding corresponging page in pagecache locked, and if
	// we would try to update that same page in pagecache it would result
	// in deadlock inside kernel.
	//
	// .loading cleanup is done once we are finished with putting the data into OS pagecache.
	// If we do it earlier - a simultaneous read covered by the same block could result
	// into missing both kernel pagecache (if not yet updated) and empty .loading[blk],
	// and thus would trigger DB access again.
	go func() {
		// XXX locking - invalidation must make sure this workers are finished.

		// XXX if direct-io: don't touch pagecache
		st := gfsconn.FileNotifyStoreCache(bfdata.Inode(), blk*zbf.blksize, blkdata)

		bfdata.loadMu.Lock()
		delete(bfdata.loading, blk)
		bfdata.loadMu.Unlock()

		if st == fuse.OK {
			return
		}

		// pagecache update failed, but it must not (we verified on startup that
		// pagecache control is supported by kernel). We can correctly live on
		// with the error, but data access will be likely very slow. Tell user
		// about the problem.
		log.Errorf("BUG: bigfile %s: blk %d: -> pagecache: %s  (ignoring, but reading from bigfile will be very slow)", zbf.POid(), blk, st)
	}()

	return nil
}


// /bigfile/<bigfileX>/head/at -> readAt serves read.
func (bf *BigFile) readAt() []byte {
	// XXX locking
	// XXX zconn -> zbf.PJar() ?
	return []byte(bf.zconn.At().String())
}





// zodbCacheControl implements zodb.LiveCacheControl to tune ZODB to never evict
// LOBTree/LOBucket from live cache. We want to keep LOBTree/LOBucket always alive
// becuse it is essentially the index where to find ZBigFile data.
//
// For the data itself - we put it to kernel pagecache and always deactivate
// from ZODB right after that.
//
// See "*/head/data invariant" in "wcfs organization" overview.
//
// TODO set it to Connection.CacheControl
type zodbCacheControl struct {}

func (cc *zodbCacheControl) WantEvict(obj zodb.IPersistent) bool {
	switch obj.(type) {
	default:
		return true

	case *btree.LOBTree:
	case *btree.LOBucket:

	// ZBlk* are kept referenced by a LOBucket, so they don't go away from Connection.objtab

	// XXX + ZBigFile ?
	}

	return false
}

// FIXME gfsconn is tmp workaround for lack of way to retrieve FileSystemConnector from nodefs.Inode
// TODO:
//	- Inode += .Mount() -> nodefs.Mount
//	- Mount:
//		.Root()		-> root Inode of the fs
//		.Connector()	-> FileSystemConnector through which fs is mounted
var gfsconn *nodefs.FileSystemConnector

func main() {
	stdlog.SetPrefix("wcfs: ")
	//log.CopyStandardLogTo("WARNING") // XXX -> "DEBUG" if -d ?
	defer log.Flush()

	debug := flag.Bool("d", false, "debug")
	autoexit := flag.Bool("autoexit", false, "automatically stop service when there is no client activity")
	// XXX option to prevent starting if wcfs was already started ?

	flag.Parse()
	if len(flag.Args()) != 2 {
		log.Fatalf("Usage: %s [OPTIONS] zurl mntpt", os.Args[0])
	}
	zurl := flag.Args()[0]
	mntpt := flag.Args()[1]

	// debug -> precise t, no dates	(XXX -> always precise t?)
	if *debug {
		stdlog.SetFlags(stdlog.Lmicroseconds)
	}

	// open zodb storage
	ctx := context.Background()	// XXX + timeout?
	zstor, err := zodb.OpenStorage(ctx, zurl, &zodb.OpenOptions{ReadOnly: true})
	if err != nil {
		log.Fatal(err)
	}
	defer zstor.Close()

	// mount root
	opts := &fuse.MountOptions{
		FsName: zurl,
		Name:   "wcfs",

		DisableXAttrs: true, // we don't use
		Debug:         *debug,
	}

	root := nodefs.NewDefaultNode()
	fssrv, fsconn, err := mount(mntpt, root, opts)
	if err != nil {
		log.Fatal(err)
	}
	gfsconn = fsconn	// FIXME temp workaround (see ^^^)

	// we require proper pagecache control (added to Linux 2.6.36 in 2010)
	supports := fssrv.KernelSettings().SupportsNotify
	if !(supports(fuse.NOTIFY_STORE_CACHE) && supports(fuse.NOTIFY_RETRIEVE_CACHE)) {
		log.Fatalf("kernel FUSE does not support pagecache control")
	}

	// add entries to /
	mkfile(root, ".wcfs", NewStaticFile([]byte(zurl)))
	mkdir(root, "bigfile", &BigFileRoot{
		Node:  nodefs.NewDefaultNode(),
		zstor: zstor,
		tab:   make(map[zodb.Oid]*BigFileDir),
	})

	// TODO handle autoexit
	_ = autoexit

	// serve client requests
	fssrv.Serve()	// XXX Serve returns no error
}
